part of '../flutterx_live_data.dart';

/// A [LiveData] mixin which keeps track of each modification to it's value and allows to undo or redo changes
///
/// Note: in order to use this LiveData, the value must be serialized and deserialized.
/// This is done by [encode] and [decode] functions
///
/// To know you can undo/redo the history you can query [canUndo] and [canRedo] value
/// Once you are sure you can execute a certain action you can call [historyUndo] and [historyRedo] methods to
/// undo/redo a change made to this LiveData, or you can call anytime [historyClear] to completely clear the history
///
/// You also can set [historyCapacity] field which determines the max size of the history (by default is 20)
/// If you set a length that is less than the current history queue the first elements of the history will be
/// removed until the history length matches the desired value
mixin HistoryMixin<T> on LiveData<T> {
  /// Queue of all saved states of this LiveData
  @protected
  final Queue<String> queue = Queue();
  int _capacity = 20;
  int _index = -1;
  bool _restoring = false;

  /// Tells if you can safely call [historyUndo].
  final LiveData<bool> canUndo = MutableLiveData(initialValue: false);

  /// Tells if you can safely call [historyRedo].
  final LiveData<bool> canRedo = MutableLiveData(initialValue: false);

  /// Get max length of history queue
  int get historyCapacity => _capacity;

  /// Set max length of history queue and automatically resize history to match desired length
  set historyCapacity(int value) {
    _capacity = value;
    while (queue.length >= value) queue.removeFirst();
  }

  /// Get current index of the history, if this index is -1 it means [queue] is empty
  int get historyIndex => _index;

  /// Set current history index, the value must be within inclusive range of -1 and [historyLength]-1
  @protected
  set historyIndex(int value) => _index = value;

  /// Get current history queue length
  int get historyLength => queue.length;

  /// Convert given value to string defaults to [jsonEncode]
  @protected
  String Function(T data)? get encode => null;

  /// Convert given string to value defaults to [jsonDecode]
  @protected
  T Function(String source)? get decode => null;

  @override
  void invalidate() {
    if (!_restoring && initialized) saveState();
    super.invalidate();
  }

  /// Reset current LiveData value to previous history value
  void historyUndo() {
    if (!canUndo._data) throw StateError('Cannot undo');
    _index--;
    restoreState();
  }

  /// Reset current LiveData value to next history value
  void historyRedo() {
    if (!canRedo._data) throw StateError('Cannot redo');
    _index++;
    restoreState();
  }

  /// Clear history queue. This will not alter current LiveData value
  void historyClear() {
    queue.clear();
    _index = -1;
    updateValues();
  }

  /// Save current state into queue
  @protected
  @mustCallSuper
  void saveState() {
    assert(initialized, 'saveState called before data has been initialized');
    final state = (encode ?? jsonEncode)(_data);
    while (queue.length > _index + 1) queue.removeLast();
    if (queue.isEmpty || state != queue.last) {
      queue.add(state);
      _index + 1 == _capacity ? queue.removeFirst() : _index++;
      updateValues();
    }
  }

  /// Restore previously saved state
  @protected
  @mustCallSuper
  void restoreState() {
    final state = (decode ?? jsonDecode)(queue.elementAt(_index));
    try {
      _restoring = true;
      setData(state);
    } finally {
      _restoring = false;
    }
    updateValues();
  }

  /// Update [canUndo] and [canRedo] fields
  @protected
  void updateValues() {
    canUndo.setData(_index > 0);
    canRedo.setData(_index < queue.length - 1);
  }
}
