part of '../flutterx_live_data.dart';

/// Link [LiveData] observers to the lifecycle of your [State]
///
/// Implement this mixin on your widget's [State] to add [Observer]s to a [LiveData] within the state lifecycle
/// Every LiveData can be observed inside [registerObservers] method that you can override
/// The observer must be added through proper extension methods and will be removed automatically on state [dispose]
/// For better performances do data.observe(_myCallbackFunction) instead of data.observe((value) => ...)
/// Check [StateObserverExt]
mixin StateObserver<T extends StatefulWidget> on State<T> {
  final Map<Object, Disposable> _disposables = {};
  Set<Object>? _keys;
  bool _dirty = false;

  @override
  void initState() {
    super.initState();
    doRegister();
  }

  @override
  void didUpdateWidget(T oldWidget) {
    super.didUpdateWidget(oldWidget);
    doRegister();
  }

  @override
  void dispose() {
    doUnregister();
    super.dispose();
  }

  /// Override this method to observe you LiveData(s)
  @protected
  FutureOr<void> registerObservers();

  /// Trigger calling [registerObservers] and automatically manage registering/disposing observers
  @protected
  void doRegister() {
    if (_keys != null) {
      _dirty = true;
      return;
    }
    final keys = _keys = _disposables.keys.toSet();
    final result = registerObservers();
    void doComplete() {
      for (final key in keys) _disposables.remove(key)!.dispose();
      _keys = null;
      if (_dirty) {
        _dirty = false;
        doRegister();
      }
    }

    result is Future ? result.whenComplete(doComplete) : doComplete();
  }

  /// Dispose all active observers
  @protected
  void doUnregister() {
    for (final key in _disposables.keys.toList(growable: false)) _disposables.remove(key)!.dispose();
  }

  void _observeAll(
      List<LiveData> sources, Function srcOnChanged, void Function(List values) onChanged, Dispatcher? dispatcher) {
    final keys = _keys;
    if (keys == null) throw StateError('observe can only be called within registerObservers function');
    final key = _CombinedObserverKey(sources, srcOnChanged);
    _disposables.containsKey(key)
        ? keys.remove(key)
        : _disposables[key] = observeAllForever(sources, onChanged, dispatcher: dispatcher);
  }

  void _observe<V>(LiveData<V> data, Observer<V> onChanged, Dispatcher? dispatcher) {
    final keys = _keys;
    if (keys == null) throw StateError('observe can only be called within registerObservers function');
    final key = _ObserverKey(data, onChanged);
    _disposables.containsKey(key)
        ? keys.remove(key)
        : _disposables[key] = data.observeForever(onChanged, dispatcher: dispatcher);
  }
}

@immutable
class _ObserverKey {
  final LiveData _data;
  final Function _onChanged;

  const _ObserverKey(this._data, this._onChanged);

  @override
  bool operator ==(Object other) =>
      other is _ObserverKey && identical(other._data, _data) && other._onChanged == _onChanged;

  @override
  int get hashCode {
    var result = 0;
    result = _combine(result, identityHashCode(_data));
    result = _combine(result, _onChanged.hashCode);
    return _finish(result);
  }
}

@immutable
class _CombinedObserverKey {
  final List<LiveData> _data;
  final Function _onChanged;

  const _CombinedObserverKey(this._data, this._onChanged);

  @override
  bool operator ==(Object other) =>
      other is _CombinedObserverKey && _listIdentical(other._data, _data) && other._onChanged == _onChanged;

  @override
  int get hashCode {
    var result = 0;
    for (final source in _data) result = _combine(result, identityHashCode(source));
    result = _combine(result, _onChanged.hashCode);
    return _finish(result);
  }
}

int _combine(int hash, int value) {
  hash = 0x1fffffff & (hash + value);
  hash = 0x1fffffff & (hash + ((0x0007ffff & hash) << 10));
  return hash ^ (hash >> 6);
}

int _finish(int hash) {
  hash = 0x1fffffff & (hash + ((0x03ffffff & hash) << 3));
  hash = hash ^ (hash >> 11);
  return 0x1fffffff & (hash + ((0x00003fff & hash) << 15));
}

bool _listIdentical(List a, List b) {
  if (a.length != b.length) return false;
  if (identical(a, b)) return true;
  for (var index = 0; index < a.length; index++) if (!identical(a[index], b[index])) return false;
  return true;
}
